print("import: 'pybamm'")
import pybamm

print("import: 'pybamm.discretisations'")
import pybamm.discretisations

print("import: 'pybamm.experiments'")
import pybamm.experiments

print("import: 'pybamm.expression_tree'")
import pybamm.expression_tree

print("import: 'pybamm.expression_tree.operations'")
import pybamm.expression_tree.operations

print("import: 'pybamm.expression_tree.printing'")
import pybamm.expression_tree.printing

print("import: 'pybamm.geometry'")
import pybamm.geometry

print("import: 'pybamm.meshes'")
import pybamm.meshes

print("import: 'pybamm.models'")
import pybamm.models

print("import: 'pybamm.models.full_battery_models'")
import pybamm.models.full_battery_models

print("import: 'pybamm.models.full_battery_models.lead_acid'")
import pybamm.models.full_battery_models.lead_acid

print("import: 'pybamm.models.full_battery_models.lithium_ion'")
import pybamm.models.full_battery_models.lithium_ion

print("import: 'pybamm.models.submodels'")
import pybamm.models.submodels

print("import: 'pybamm.models.submodels.active_material'")
import pybamm.models.submodels.active_material

print("import: 'pybamm.models.submodels.convection'")
import pybamm.models.submodels.convection

print("import: 'pybamm.models.submodels.convection.through_cell'")
import pybamm.models.submodels.convection.through_cell

print("import: 'pybamm.models.submodels.convection.transverse'")
import pybamm.models.submodels.convection.transverse

print("import: 'pybamm.models.submodels.current_collector'")
import pybamm.models.submodels.current_collector

print("import: 'pybamm.models.submodels.electrode'")
import pybamm.models.submodels.electrode

print("import: 'pybamm.models.submodels.electrode.ohm'")
import pybamm.models.submodels.electrode.ohm

print("import: 'pybamm.models.submodels.electrolyte_conductivity'")
import pybamm.models.submodels.electrolyte_conductivity

print("import: 'pybamm.models.submodels.electrolyte_conductivity.surface_potential_form'")
import pybamm.models.submodels.electrolyte_conductivity.surface_potential_form

print("import: 'pybamm.models.submodels.electrolyte_diffusion'")
import pybamm.models.submodels.electrolyte_diffusion

print("import: 'pybamm.models.submodels.external_circuit'")
import pybamm.models.submodels.external_circuit

print("import: 'pybamm.models.submodels.interface'")
import pybamm.models.submodels.interface

print("import: 'pybamm.models.submodels.interface.interface_utilisation'")
import pybamm.models.submodels.interface.interface_utilisation

print("import: 'pybamm.models.submodels.interface.kinetics'")
import pybamm.models.submodels.interface.kinetics

print("import: 'pybamm.models.submodels.interface.kinetics.first_order_kinetics'")
import pybamm.models.submodels.interface.kinetics.first_order_kinetics

print("import: 'pybamm.models.submodels.interface.kinetics.inverse_kinetics'")
import pybamm.models.submodels.interface.kinetics.inverse_kinetics

print("import: 'pybamm.models.submodels.interface.lithium_plating'")
import pybamm.models.submodels.interface.lithium_plating

print("import: 'pybamm.models.submodels.interface.sei'")
import pybamm.models.submodels.interface.sei

print("import: 'pybamm.models.submodels.oxygen_diffusion'")
import pybamm.models.submodels.oxygen_diffusion

print("import: 'pybamm.models.submodels.particle'")
import pybamm.models.submodels.particle

print("import: 'pybamm.models.submodels.particle.no_distribution'")
import pybamm.models.submodels.particle.no_distribution

print("import: 'pybamm.models.submodels.particle.size_distribution'")
import pybamm.models.submodels.particle.size_distribution

print("import: 'pybamm.models.submodels.particle_mechanics'")
import pybamm.models.submodels.particle_mechanics

print("import: 'pybamm.models.submodels.porosity'")
import pybamm.models.submodels.porosity

print("import: 'pybamm.models.submodels.thermal'")
import pybamm.models.submodels.thermal

print("import: 'pybamm.models.submodels.thermal.pouch_cell'")
import pybamm.models.submodels.thermal.pouch_cell

print("import: 'pybamm.models.submodels.tortuosity'")
import pybamm.models.submodels.tortuosity

print("import: 'pybamm.parameters'")
import pybamm.parameters

print("import: 'pybamm.plotting'")
import pybamm.plotting

print("import: 'pybamm.solvers'")
import pybamm.solvers

print("import: 'pybamm.solvers.c_solvers'")
import pybamm.solvers.c_solvers

print("import: 'pybamm.spatial_methods'")
import pybamm.spatial_methods

